#include "qmessagebox.h"

#include "GLPlot.h"


GLPlot::GLPlot(QWidget* parent) : QGLWidget(parent) {
  drawMode = 0;
  vAverage = NULL;
  for( int i=0; i<GLP_NUM_DRAW_MODES; i++ ) {
    scale[i].set(1, 1, 1);
  }
  setLightPos(60);
  light = true;
  wireframe = false;
  useImage = false;
  showInput = false;
  showGrid = true;
  overlay = NULL;
  oscData = NULL;
  resizeOscilloscope(500);
  dTime.start();
  pxPerUnit = 1.0;
  drawEnabled = true;
  storeEnabled = false;
}

void GLPlot::resizeOscilloscope(int newSize) {
  oscPoint* oldData = oscData;
  oscPoint* newData = new oscPoint[newSize];
  for( int i=0; i<newSize; i++ ) {
    newData[i].val = 0.0;
    newData[i].time = 0;
  }
  oscData = newData;
  oscPointer = 0;
  oscLength = newSize;
  if( oldData != NULL ) {
    delete [] oldData;
  }
}

void GLPlot::doVertex(Vector v) {
  glVertex3f(v[0], v[1], v[2]);
}

void GLPlot::setPPU(float p) {
  changeScale(0, p/pxPerUnit, 1.0);
  pxPerUnit = p;
}

void GLPlot::changeScale(int n, float dx, float dy) {
      plotDataMutex.lock();
      Vector s(dx, dy, 1.0);
      Matrix m = scaleMatrix(s);
      scale[n] = m * scale[n];
      translate[n] = translateMatrix(Vector(0.5, 0.5)) * m * translateMatrix(Vector(-0.5, -0.5)) * translate[n];
      plotDataMutex.unlock();
}

void GLPlot::init2D() {
  glBlendFunc (GL_SRC_ALPHA, GL_DST_ALPHA);
  glEnable (GL_BLEND);
  glEnable (GL_ALPHA_TEST);
  glAlphaFunc (GL_ALWAYS, 0.5);
  glEnable(GL_ALPHA);
  glDisable(GL_DEPTH_TEST);
}

void GLPlot::init3D() {
  glDisable (GL_BLEND);
  glDisable (GL_ALPHA_TEST);
  glDisable(GL_ALPHA);
//  glEnable(GL_POLYGON_SMOOTH);
  glEnable(GL_DEPTH_TEST);
  glDepthFunc(GL_LESS);
  glCullFace(GL_NONE);

  // Lighting
  float white[] = {1.0, 1.0, 1.0, 1.0};
  float gray[] = {0.6, 0.6, 0.6, 1.0};
  float gray2[] = {0.25, 0.25, 0.25, 1.0};
  glLightfv(GL_LIGHT0, GL_SPECULAR, white);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, gray);
  glLightfv(GL_LIGHT0, GL_AMBIENT, gray2);
}

void GLPlot::initializeGL() {
  glClearColor( 0.0, 0.0, 0.0, 0.0 );
}

void GLPlot::resizeGL(int w, int h) {
  glViewport( 0, 0, (GLint)w, (GLint)h );
  cam.aspect = (float)h/(float)w;
}

void GLPlot::paintGL() {
/*  if( dTime.elapsed() < 50 ) {
    return;
  }
  else {
    dTime.restart();
  }*/
  plotDataMutex.lock();
  glDisable(GL_LIGHTING);
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  
  
  if( showInput ) {
    drawInput();
  }
    
  if( drawEnabled ) {
    if( drawMode == 0 ) {
      init2D();
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glTranslatef(translate[drawMode][0], translate[drawMode][1], translate[drawMode][2]);
      glScalef(scale[drawMode][0], scale[drawMode][1], scale[drawMode][2]);
      
      if( showGrid )
        drawGrid();
      drawPlot();
      drawCrosshair();
    }
    else if( drawMode == 1 ) {
      init3D();
      if( light ) {
        setupLighting();
      }
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glTranslatef(translate[drawMode][0], translate[drawMode][1], translate[drawMode][2]);
      
      cam.setProjMatrix();
      
      if( wireframe) {
        drawWireframe();
      }
      else {
        drawSurface();
      }
    }
    else if( drawMode == 2 ) {
      init2D();
      
      glMatrixMode(GL_MODELVIEW);
      glLoadIdentity();
      glTranslatef(translate[drawMode][0], translate[drawMode][1], translate[drawMode][2]);
      glScalef(scale[drawMode][0], scale[drawMode][1], scale[drawMode][2]);
      
      if( showGrid )
        drawGrid();
      drawScope();
      drawCrosshair();
    }
  }
  plotDataMutex.unlock();
}


void GLPlot::newFrame(Frame newFrame) {
  plotDataMutex.lock();
  if( plotData.buf != NULL ) {
    delete [] plotData.buf;
  }
  plotData = newFrame;
  if( drawMode == 0 ) {
    recomputeAverage();
  }
  else if( drawMode == 2 ) {
    refreshOscilloscope();
  }
  if( storeEnabled ) {
    savePlot( storeFile, true );
  }
  plotDataMutex.unlock();
  update();
}





Camera::Camera() {
  phi = -60;
  th = 0;
  dist = 200;
  fov = 70;
  aspect = 1.0;
}

void Camera::setProjMatrix() {
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  float near = 1;
  float w2 = near * tan(fov/2.0);
  float h2 = w2 * aspect;
  glFrustum(-w2, w2, -h2, h2, near, 10000);
  glTranslatef(0, 0, -dist);
  glRotatef(phi, 1, 0, 0);
  glRotatef(th, 0, 0, 1);
}


void GLPlot::mouseMoveEvent(QMouseEvent * e) {
  Vector m = Vector(e->x(), e->y());
  Vector d = m - mousePos;
  
  if( drawMode == 0 || drawMode == 2 ) {
    if( mouseButton == Qt::LeftButton ) {
      plotDataMutex.lock();
      Vector t(d[0]/(float)width(), -d[1]/(float)height());
      translate[drawMode] += t;
      plotDataMutex.unlock();
    }
    else if( mouseButton == Qt::RightButton ) {
      changeScale(drawMode, pow(1.01, d[0]), pow(1.01, -d[1]));
/*      plotDataMutex.lock();
      Vector s(pow(1.01, d[0]), pow(1.01, -d[1]), 1.0);
      Matrix m = scaleMatrix(s);
      scale[drawMode] = m * scale[drawMode];
      translate[drawMode] = translateMatrix(Vector(0.5, 0.5)) * m * translateMatrix(Vector(-0.5, -0.5)) * translate[drawMode];
      plotDataMutex.unlock();*/
    }
    else if( mouseButton == Qt::MidButton ) {
    }
  }
  else if( drawMode == 1 ) {
    if( mouseButton == Qt::LeftButton ) {
      cam.th += d[0]/ 3.0;
      cam.phi += d[1]/ 3.0;
    }
    else if( mouseButton == Qt::RightButton ) {
      scale[drawMode][2] *= 1.0 - (d[1] / 50.0);
      if( scale[drawMode][2] < 0.00001 ) {
        scale[drawMode][2] = 0.00001;
      }
    }
    else if( mouseButton == Qt::MidButton ) {
      translate[drawMode][2] -= d[1] / 3.0;
    }
  }
  
  
  mousePos.set(m);
  update();
}

void GLPlot::mousePressEvent(QMouseEvent * e) {
  mousePress.set(e->x(), e->y());
  mousePos.set(e->x(), e->y());
  mouseButton = e->button();
//   printf("pressed button %d\n", e->button());
  update();
}

void GLPlot::mouseReleaseEvent(QMouseEvent * e) {
  mouseButton = 0;
  update();
}

void GLPlot::wheelEvent(QWheelEvent* e) {
  float d = e->delta();
  cam.dist *= 1.0 - (d/1200.0);
  update();
}

void GLPlot::setDrawMode(int i) {
  drawMode = i;
  update();
}

void GLPlot::setShowInput(bool b) {
  showInput = b;
  update();
}

void GLPlot::setLight(bool b) {
  light = b;
  update();
}

void GLPlot::setUseImage(bool b) {
  useImage = b;
//   printf("useImage = %s\n", useImage? "true" : "false");
  update();
}

void GLPlot::setWireframe(bool b) {
  wireframe = b;
  update();
}

void GLPlot::setLightPos(float angle) {
  lightPos = Vector(-cam.dist * sin(angle * D2R), 0, cam.dist * cos(angle * D2R));
}


void GLPlot::drawInput() {
  glDisable(GL_LIGHTING);
  glMatrixMode(GL_MODELVIEW);
  glLoadIdentity();
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, width()-1, height()-1, 0, -1, 1);
    
//     glDrawPixels(plotData.width, plotData.height, GL_LUMINANCE, GL_FLOAT, plotData.buf);
  glPointSize(1);
  glBegin(GL_POINTS);
  int index = 0;
  for( int i=0; i<plotData.height; i++ ) {
    for( int j=0; j<plotData.width; j++ ) {
      float c = plotData.buf[index];
      glColor4f(c, c, c, 1);
      glVertex2f(j, i);
      index++;
    }
  }
  glEnd();
}


void GLPlot::recomputeAverage() {
      // Run average
  delete [] vAverage;
    
  int index = 0;
  vAverage = new float[plotData.width];
  float s = 1.0 / (float)plotData.height;
  for( int i=0; i<plotData.width; i++ ) {
    vAverage[i] = 0.0;
  }
  for( int j=0; j<plotData.height; j++ ) {
    for( int i=0; i<plotData.width; i++ ) {
      vAverage[i] += plotData.buf[index];
      index++;
    }
  }
  for( int i=0; i<plotData.width; i++ ) {
    vAverage[i] *= s;
  }
}

void GLPlot::refreshOscilloscope() {
  float avg = 0.0;
  int c = 0;
  for( int j=0; j<plotData.height; j++ ) {
    for( int i=0; i<plotData.width; i++ ) {
      avg += plotData.buf[c];
      c++;
    }
  }
  avg /= (float)c;
  oscPointer = (oscPointer+1) % oscLength;
  oscData[oscPointer].val = avg;
  oscData[oscPointer].time = plotData.time;
}

void GLPlot::drawScope() {
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, 1, 0, 1, -1, 1);
    
/*  float x = 0;
  float dx = 1.0 / (oscLength - 1);*/
  int startTime = oscData[oscPointer].time;
  glColor4f(1, 1, 1, 1);
  glBegin(GL_LINE_STRIP);
  for( int i=0; i<oscLength; i++ ) {
    int ind = (i+oscPointer+1)%oscLength;
    glVertex2f((((float)oscData[ind].time-startTime)/1000.0)+1, oscData[ind].val);
//     x += dx;
  }
  glEnd();
}

void GLPlot::drawGrid() {
  
  if( drawMode == 1 ) {
  
  }
  else {
    Matrix m = scaleMatrix(scale[drawMode]).inverse() * translateMatrix(translate[drawMode]).inverse();
    Vector bl = m * Vector();
    Vector ur = m * Vector(1,1);
    Vector d = ur-bl;
    
    // Draw grid lines for x and y axes
    float l10 = 1.0 / log(10.0);
    glBegin(GL_LINES);
    for( int ax = 0; ax<2; ax++ ) {
      int bx = (ax+1) % 2;
      int p = floor((log(d[ax]) * l10));
      // draw the 2 most visible scales
//       printf("ax %d: dx = %f, pow = %d - %d\n", ax, d[ax], p-2, p-1);
      for( int scale = p; scale>p-2; scale-- ) {
        float dx = pow(10.0, scale);
        int num = (int)((d[ax] / dx)+2);
        // draw each line
        float start = dx * (float)((int)(bl[ax] / dx));
        Vector sv, ev;
        sv[ax] = start;
        sv[bx] = bl[bx];
        ev[ax] = start;
        ev[bx] = ur[bx];
        float alpha;
        if( ax == 0 )
          alpha = 0.5 - (float)((num*cam.aspect)-10)/80;
        else
          alpha = 0.5 - (float)(num-10)/80;
        glColor4f(0.5, 0.5, 0.5, alpha);
        for( int n=0; n<num; n++ ) {
          doVertex(sv);
          doVertex(ev);
          sv[ax] += dx;
          ev[ax] += dx;
        }
      }
    }
    glEnd();    
  }
}

void GLPlot::drawCrosshair() {
    init2D();
    glMatrixMode(GL_PROJECTION);
    glLoadIdentity();
    glOrtho(0, 1, 0, 1, -1, 1);
    glMatrixMode(GL_MODELVIEW);
    glLoadIdentity();
    // Draw center cross  
    glColor4f(0.5, 0.5, 0.5, 0.5);
    glBegin(GL_LINES);
    glVertex2f(0.5, 0.55);  
    glVertex2f(0.5, 0.45);  
    glVertex2f(0.5 + (0.05*cam.aspect), 0.5);  
    glVertex2f(0.5 - (0.05*cam.aspect), 0.5);  
    glEnd();
}

void GLPlot::drawPlot() {
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  glOrtho(0, 1, 0, 1, -1, 1);
    
  float x = 0;
  float dx = 1.0 / pxPerUnit; //(plotData.width - 1);
  glColor4f(0.8, 1, 0.5, 1);
  glBegin(GL_LINE_STRIP);
  for( int i=0; i<plotData.width; i++ ) {
    glVertex2f(x, vAverage[i]);
    x += dx;
  }
  glEnd();
  
  // Draw overlay
  glColor4f(0.5, 1, 0.5, 1);
  glBegin(GL_LINE_STRIP);
  if( overlay != NULL ) {
    x = 0.0;
    for( int i=0; i<plotData.width; i++ ) {
      glVertex2f(x, overlay[i]);
      x += dx;
    }
  }
  glEnd();
}


void GLPlot::setupLighting() {
  if( light ) {
    float light_position[] = {lightPos[0], lightPos[1], lightPos[2], 0};
    glLightfv(GL_LIGHT0, GL_POSITION, light_position);
    glEnable(GL_LIGHTING);
    glEnable(GL_LIGHT0);
    float gray[] = {.1, .1, .1, 1};
    float white[] = {1, 1, 1, 1};
    glMaterialfv(GL_FRONT, GL_AMBIENT, gray);
    glMaterialfv(GL_FRONT, GL_DIFFUSE, white);
  }
  else {
    glDisable(GL_LIGHTING);
  }
}

void GLPlot::drawSurface() {
  int w = plotData.width;
  int h = plotData.height;
  float w2 = (float)w / 2.0;
  float h2 = (float)h / 2.0;
  float cont = pow(1.1, plotData.contrast);
  float brt = plotData.brightness;
//   printf("cont: %f brt: %f\n", cont, brt);
  
  
  glBegin(GL_QUADS);
  for( int i=0; i<w-1; i++ ) {
    for( int j=0; j<h-1; j++ ) {
      float c[4];
      c[0] = plotData.val(i, j);
      c[1] = plotData.val(i+1, j);
      c[2] = plotData.val(i, j+1);
      c[3] = plotData.val(i+1, j+1);
      
      Vector v[4];
      v[0].set(i-w2, h2-j, c[0]*scale[drawMode][2]);
      v[1].set(i+1-w2, h2-j, c[1]*scale[drawMode][2]);
      v[2].set(i-w2, h2-j-1, c[2]*scale[drawMode][2]);
      v[3].set(i+1-w2, h2-j-1, c[3]*scale[drawMode][2]);
      
      Vector n = ((v[1]-v[2]) * (v[0]-v[3])).norm();
            
      if( ! useImage ) {
        glColor4f(1, 1, 1, 1);
        glNormal3f(n[0], n[1], n[2]);
        doVertex(v[0]);
        doVertex(v[1]);
        doVertex(v[3]);
        doVertex(v[2]);
      }
      else {
        for( int i=0; i<4; i++ ) {
          c[i] *= cont;
          c[i] += brt;
        }
        
        if( light ) {
          glNormal3f(n[0], n[1], n[2]);
          float color[] = {c[0], c[0], c[0], 1};
          glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
          doVertex(v[0]);
          float color1[] = {c[1], c[1], c[1], 1};
          glMaterialfv(GL_FRONT, GL_DIFFUSE, color1);
          doVertex(v[1]);
          float color3[] = {c[3], c[3], c[3], 1};
          glMaterialfv(GL_FRONT, GL_DIFFUSE, color3);
          doVertex(v[3]);
          float color2[] = {c[2], c[2], c[2], 1};
          glMaterialfv(GL_FRONT, GL_DIFFUSE, color2);
          doVertex(v[2]);
        }
        else {
          glNormal3f(n[0], n[1], n[2]);
          glColor4f(c[0], c[0], c[0], 1);
          doVertex(v[0]);
          glColor4f(c[1], c[1], c[1], 1);
          doVertex(v[1]);
          glColor4f(c[3], c[3], c[3], 1);
          doVertex(v[3]);
          glColor4f(c[2], c[2], c[2], 1);
          doVertex(v[2]);
        }
      }
      
    }
  }
  glEnd();
}


void GLPlot::drawWireframe() {
  int w = plotData.width;
  int h = plotData.height;
  float w2 = (float)w / 2.0;
  float h2 = (float)h / 2.0;
  glColor4f(1, 1, 1, 1);
  for( int i=0; i<w-1; i++ ) {
    glBegin(GL_LINE_STRIP);
    for( int j=0; j<h-1; j++ ) {
      float c;
      c = plotData.val(i,   j);
      Vector v = Vector(-w2 + i, h2 - j, c*scale[drawMode][2]);
      if( ! useImage ) {
        glColor4f(1, 1, 1, 1);
      }
      else {
        glColor4f(c, c, c, 1);
      }
      doVertex(v);
    }
    glEnd();
  }
  for( int j=0; j<h-1; j++ ) {
    glBegin(GL_LINE_STRIP);
    for( int i=0; i<w-1; i++ ) {
      float c;
      c = plotData.val(i,   j);
      Vector v = Vector(-w2 + i, h2 - j, c*scale[drawMode][2]);
      if( ! useImage ) {
        glColor4f(1, 1, 1, 1);
      }
      else {
        glColor4f(c, c, c, 1);
      }
      doVertex(v);
    }
    glEnd();
  }
}

void GLPlot::savePlot(QString filename, bool stream) {
  
  FILE* fd;
  if( stream ) {
    fd = fopen(filename, "a");
  }
  else {
    fd = fopen(filename, "w");
  }
  if( stream ) {
    fprintf(fd, "%d,%d,%d,\n", plotData.time, plotData.width, plotData.height);
  }
    
  if( fd == NULL ) {
    QMessageBox::warning( this, "Bummer, man.", "Could not open file for writing!", QMessageBox::Ok, QMessageBox::NoButton, QMessageBox::NoButton);
    return;
  }
  if( drawMode == 0 ) {
    float upp = 1.0 / pxPerUnit;
    for( int i=0; i<plotData.width; i++ ) {
      fprintf(fd, "%f,%f\n", ((float)i)*upp, vAverage[i]);
    }
  }
  else if( drawMode == 1 ) {
    int index = 0;
    for( int j=0; j<plotData.height; j++ ) {
      for( int i=0; i<plotData.width; i++ ) {
        fprintf(fd, "%f", plotData.buf[index]);
        if( i == plotData.width - 1 )
          fprintf(fd, "\n");
        else 
          fprintf(fd, ",");
          
        index++;
      }
    }
  }
  else if( drawMode == 2 ) {
    int t = oscData[oscPointer].time;
    if( stream ) {
      int i = oscLength-1;
      int j = (i+oscPointer+1) % oscLength;
      fprintf(fd, "%f, %f\n", (float)(oscData[j].time-t) / 1000, oscData[j].val);
    }
    else {
      for( int i=0; i<oscLength; i++ ) {
        int j = (i+oscPointer+1) % oscLength;
        fprintf(fd, "%f, %f\n", (float)(oscData[j].time-t) / 1000, oscData[j].val);
      }
    }
  }
  fclose(fd);
}

float* GLPlot::getPlotAvg(int* width) {
  float* data;
  if( drawMode != 0 ) {
    return NULL;
  }
  else {
    data = new float[plotData.width];
    for( int i=0; i<plotData.width; i++ ) {
      data[i] = vAverage[i];
    }
  }
  *width = plotData.width;
  return data;
}

void GLPlot::draw2DOverlay(float* data, int width) {
  delete [] overlay;
  overlay = data;
}

void GLPlot::setDrawEnabled(bool s) {
  drawEnabled = s;
}

void GLPlot::setStoreEnabled(bool s) {
  if( s && ! storeEnabled ) {
    storeFile = QFileDialog::getSaveFileName(
      "", 
      "Comma Separated Values (*.csv)", 
      this, 
      "save file dialog", 
      "Choose a filename to save under"
    );
  }
  storeEnabled = s;
}

void GLPlot::setShowGrid(bool s) {
  showGrid = s;
}