#include "PlotThread.h"

PlotThread::PlotThread(GLPlot* p) {
  plotWidget = p;
  interpolate = true;
  running = false;
  nextRegionReady = false;
  downsample = 1;
}

void PlotThread::setDownsample( int ds ) {
  downsample = ds;
}

void PlotThread::run() {
//   printf("Plot thread running..\n");
  running = true;
  while( running ) {
    if( claimNextFrame() || readNewRegion() ) {
      // something has changed, generate a new buffer to send to the plot widget
      Frame newPlotFrame = makeFrame();
      if( plotWidget != NULL ) {
        plotWidget->newFrame(newPlotFrame);
      }
    }
    else {
      QThread::usleep(10000);
    }
  }
}

void PlotThread::stop() {
  running = false;
}

void PlotThread::setRegion( RectRegion newReg ) {
  regionMutex.lock();
  nextRegion = newReg;
  nextRegionReady = true;
  regionMutex.unlock();
}

bool PlotThread::readNewRegion() {
  bool ret;
  
  regionMutex.lock();
  if( nextRegionReady ) {
    currentRegion = nextRegion;
    ret = true;
//     printf("PlotThread: Got new region: %d %d %d %d %f\n", currentRegion.x, currentRegion.y, currentRegion.width, currentRegion.height, currentRegion.angle);
    nextRegionReady = false;
  }
  else {
    ret = false;
  }
  regionMutex.unlock();
  return ret;
}

Frame PlotThread::swapNewFrame(Frame newFrame) {
  Frame ret;
  if( frameMutex.tryLock() ) {
    if( nextFrame.buf == NULL ) {
      nextFrame = newFrame;
      ret = Frame((float*)NULL, 0, 0, 0, 0);
//      printf("PlotThread: used last frame; returning null frame\n");
    }
    else {
      ret = nextFrame;
      nextFrame = newFrame;
//      printf("PlotThread: returning unused frame\n");
    }
    frameMutex.unlock();
    return ret;
  }
  else {
    // Could not lock mutex; just reuse the same frame
/*    printf("PlotThread: could not lock mutex; returning same frame\n");*/
    return newFrame;
  }
}

bool PlotThread::claimNextFrame() {
  bool ret;
  frameMutex.lock();
  
  if( nextFrame.buf == NULL ) {
/*    printf("PlotThread: no new frames waiting.\n");*/
    ret = false;
  }
  else {
/*    printf("PlotThread: claiming frame\n");*/
    // delete old frame
    delete [] currentFrame.buf;
  
    // lay claim to the next frame
    currentFrame = nextFrame;
  
    // mark the old one as used
    nextFrame.buf = NULL;
    ret = true;
  }
  frameMutex.unlock();
  return ret;
}


Frame PlotThread::makeFrame() {
  int ds = downsample;
  int w, h;
  w = currentRegion.width / ds;
  h = currentRegion.height / ds;
  Frame tmp(w, h, currentFrame.time, currentFrame.vpp);
  tmp.contrast = currentFrame.contrast;
  tmp.brightness = currentFrame.brightness;
  float dx = cos(currentRegion.angle);
  float dy = sin(currentRegion.angle);
  float x, y;
  float startx = currentRegion.x;
  float starty = currentRegion.y;
  int index = 0;
  for( int i0=0; i0+ds-1<currentRegion.height; i0+=ds ) {
    for( int j0=0; j0+ds-1<currentRegion.width; j0+=ds ) {
      tmp.buf[index] = 0.0;
      for( int i1=0; i1<ds; i1++ ) {
        for( int j1=0; j1<ds; j1++ ) {
          int i = i0 + i1;
          int j = j0 + j1;
          x = startx + j*dx + i*dy;
          y = starty - j*dy + i*dx;
          if( y >= 0 && y < currentFrame.height && x >= 0 && x < currentFrame.width ) {
            if( interpolate ) {
              float xs = x - (float)(int)x;
              float ys = y - (float)(int)y;
              float rxs = 1.0 - xs;
              float rys = 1.0 - ys;
              float v1 = currentFrame.val(x,   y);
              float v2 = currentFrame.val(x+1, y);
              float v3 = currentFrame.val(x,   y+1);
              float v4 = currentFrame.val(x+1, y+1);
              tmp.buf[index] += v1 * rxs * rys + v2 * xs * rys + v3 * rxs * ys + v4 * xs * ys;
            }
            else {
              int cfindex = ((int)y * currentFrame.width) + (int)x;
              tmp.buf[index] += currentFrame.buf[cfindex];
            }
          }
          else {
            tmp.buf[index] += 0;
          }
        }
      }
      tmp.buf[index] /= (float)(ds*ds);
      index++;
    }
  }
  return tmp;
}

