// Copyright 2003,2005 Luke Campagnola <lcampagn@mines.edu>, Colorado School of Mines
// 
// This file is part of libvectormath.
// 
// libvectormath is free software; you can redistribute it and/or modify
// it under the terms of the GNU Lesser General Public License as published by
// the Free Software Foundation; either version 2.1 of the License, or
// (at your option) any later version.
// 
// libvectormath is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
// 
// You should have received a copy of the GNU Lesser General Public License
// along with libvectormath; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA



/***************************************
/ Vector and Matrix classes
/ Luke Campagnola (lcampagn@mines.edu)
/ 2003.10.15
***************************************/

#include <math.h>
#include <stdio.h>

#include "vectormath.h"


 // First some helper functions:

float min(float a, float b) {
  return a>b ? b : a;
}
float max(float a, float b) {
  return a>b ? a : b;
}
 
 
 // Now the good stuff:
 
Vector::Vector(float a, float b, float c, float d) {
  x[0]=a;
  x[1]=b;
  x[2]=c;
  x[3]=d;  
}

Vector::~Vector() {
}

float& Vector::operator[] ( int index ) {
  return x[index];
}

float Vector::operator[] ( int index ) const {
  return x[index];
}

void Vector::set(float a, float b, float c, float d) {
  x[0]=a;
  x[1]=b;
  x[2]=c;
  x[3]=d;  
  //magKnown = false;
}

void Vector::set(const Vector& a) {
  x[0] = a.x[0];
  x[1] = a.x[1];
  x[2] = a.x[2];
  x[3] = a.x[3];
  //magKnown = false;
}


// Vector-Vector operations
  
Vector Vector::operator+ (const Vector& op ) const {
  return Vector(x[0]+op.x[0], x[1]+op.x[1], x[2]+op.x[2],1);
}

Vector Vector::operator- (const Vector& op ) const {
  return Vector(x[0]-op.x[0], x[1]-op.x[1], x[2]-op.x[2],1);
}

Vector Vector::operator- () const {
  return Vector(-x[0], -x[1], -x[2],1);
}

void Vector::operator+= (const Vector& op ) {
  this->x[0]+= op.x[0];
  this->x[1]+= op.x[1];
  this->x[2]+= op.x[2];
//   magKnown = false;
}

void Vector::operator-= (const Vector& op ) {
  this->x[0]-= op.x[0];
  this->x[1]-= op.x[1];
  this->x[2]-= op.x[2];
//   magKnown = false;
}

  // Cross product  
Vector Vector::operator* (const Vector& op ) const {
  return Vector(x[1]*op.x[2]-x[2]*op.x[1], x[2]*op.x[0]-x[0]*op.x[2], x[0]*op.x[1]-x[1]*op.x[0], 1);
}

void Vector::operator*= (const Vector& op ) {
  x[0] = x[1]*op.x[2]-x[2]*op.x[1];
  x[1] = x[2]*op.x[0]-x[0]*op.x[2];
  x[2] = x[0]*op.x[1]-x[1]*op.x[0];
//   magKnown = false;
}

  // Dot product
float Vector::operator& (const Vector& op ) const {
  return (float)( x[0]*op.x[0] + x[1]*op.x[1] + x[2]*op.x[2] );
}


float Vector::dot(const Vector& op ) const {
  return (float)( x[0]*op.x[0] + x[1]*op.x[1] + x[2]*op.x[2] );
}

Vector Vector::proj( const Vector& op ) const {
  Vector bhat = op.norm();
  Vector a = bhat * dot(bhat);
  a.x[3] = 1;
  return a;
}

float Vector::angle( const Vector& op ) const {
  float angle = acos(dot(op)/(mag()*op.mag()));
  if( angle != angle )  // Check for inf values
    angle = 0;
  return(angle);
}

// Vector-scalar operations

Vector Vector::operator* ( float scl ) const {
  return Vector(scl*x[0], scl*x[1], scl*x[2], 1);
} 

Vector Vector::operator/ ( float scl ) const {
  return Vector(x[0]/scl, x[1]/scl, x[2]/scl, 1);
} 

void Vector::operator*= ( float scl ) {
  this->x[0]*= scl;
  this->x[1]*= scl;
  this->x[2]*= scl;
//   magKnown = false;
} 

void Vector::operator/= ( float scl ) {
  this->x[0]/= scl;
  this->x[1]/= scl;
  this->x[2]/= scl;
//   magKnown = false;
} 

// Vector operations

Vector Vector::norm() const {
  float len = mag();
  if( len != 0 ) {
    float id = 1/len;
    return Vector(x[0] * id, x[1] * id, x[2] * id, 1);
  }
  return Vector(*this);
}

void Vector::normalize() {
  float len = mag();
  if( len != 0 ) {
    float id = 1/len;
    x[0] *= id;
    x[1] *= id;
    x[2] *= id;
  }
}


float Vector::mag() const {
/*
  if( ! magKnown ) {
    #ifdef _AFSQRT_
      len=afsqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
    #else
      len = sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
    #endif
    magKnown = true;
  }
  return len;
  */
    #ifdef _AFSQRT_
      return afsqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
    #else
      return sqrt(x[0]*x[0]+x[1]*x[1]+x[2]*x[2]);
    #endif
}


void Vector::print() const {
  for( int i=0; i<4; i++ ) {
    printf("%f ", x[i]);
  }
  printf("\n");
}

  // Finds the point closest to two lines
Vector lineLineMidpoint(Vector s1, Vector e1, Vector s2, Vector e2) {
    // First, find the cross product of the two lines
  Vector lineToLine = (e1-s1) * (e2-s2);
    // Now rotate both lines so that they are in the x-y plane
  float angle = lineToLine.angle(Vector(0, 0, 1));
  Vector axis = lineToLine * Vector(0, 0, 1);
  Matrix r = rotateMatrix(axis, angle);
  Vector l1s = r * s1;
  Vector l1e = r * e1;
  Vector l2s = r * s2;
  Vector l2e = r * e2;
    
    // Now the midpoint is easier to calculate
    
    // m = y2-y1 / x2-x1
  float m1 = (l1e[1]-l1s[1]) / (l1e[0] - l1s[0]);
  float m2 = (l2e[1]-l2s[1]) / (l2e[0] - l2s[0]);
    // b = y - mx
  float b1 = l1s[1] - m1 * l1s[0];
  float b2 = l2s[1] - m2 * l2s[0];
  
  float x = (b2-b1) / (m1-m2);
  float y = m1 * x + b1;
  float z = (l1s[2] + l2s[2]) / 2;
  
  Vector midpoint(x, y, z);

    // now rotate the point back to the original coordinates
  return rotateMatrix(axis, -angle) * midpoint;
}





   //// Matrix Methods /////////






Matrix::Matrix() {    // initialize with identity Matrix
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      if( i==j ) {
        x[i][j] = 1;
      }
      else {
        x[i][j] = 0;
      }        
    }
  }
}

Matrix::Matrix(const double m[4][4]) {    // initialize with given Matrix
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      x[i][j] = m[i][j];
    }
  }
}
Matrix::Matrix(const float m[4][4]) {    // initialize with given Matrix
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      x[i][j] = m[i][j];
    }
  }
}

float& Matrix::operator() ( int r, int c ) {
  return x[r][c];
}

float Matrix::operator() ( int r, int c ) const {
  return x[r][c];
}


Matrix Matrix::operator* (const Matrix& a) const {
  Matrix b;
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      b.x[i][j] = 0;
      for( int k=0; k<4; k++ ) {
        b.x[i][j] += x[i][k] * a.x[k][j];
      }
    }
  }
  return b;
}

Matrix Matrix::operator+ (const Matrix& a) const {
  Matrix b;
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      b.x[i][j] = x[i][j] + a.x[i][j];
    }
  }
  return b;
}

Matrix Matrix::operator* (float a) const {
  Matrix b;
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      b.x[i][j] = x[i][j] * a;
    }
  }
  return b;
}

Matrix Matrix::operator/ (float a) const {
  Matrix b;
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      b.x[i][j] = x[i][j] / a;
    }
  }
  return b;
}

Vector Matrix::operator* (const Vector& a) const {
  Vector b;
  for( int i=0; i<4; i++ ) {
    b[i] = 0;
    for( int k=0; k<4; k++ ) {
      b[i] += x[i][k] * a[k];
    }
  }
//  b[3] = 1;
  return b;
}

float* Matrix::glMatrix() {
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      glMat[4*j+i] = x[i][j];
    }
  }
  return glMat;
}

void Matrix::print() const {
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      printf("%f ", x[i][j]);
    }
    printf("\n") ;
  }
}

float Matrix::det() const {
  return x[0][0] * subDet(0, 0) - x[0][1] * subDet(0, 1) + x[0][2] * subDet(0, 2) - x[0][3] * subDet(0, 3);
}

float Matrix::subDet(int i, int j) const {
  float d = 0;
  int sj = (j+1)%4; 
  int i1 = (i+1)%4; 
  int i2 = (i+2)%4; 
  int i3 = (i+3)%4; 
  
  for( int a=0; a<3; a++ ) {
    int j1 = (sj+((a+0)%3)) % 4;
    int j2 = (sj+((a+1)%3)) % 4;
    int j3 = (sj+((a+2)%3)) % 4;
    d += x[i1][j1] * ((x[i2][j2]*x[i3][j3])-(x[i2][j3]*x[i3][j2]));
  }
  return d;
}

Matrix Matrix::inverse() const {
  Matrix tmp;
  float id = 1.0/det();
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      tmp.x[j][i] = subDet(i, j) * id * pow(-1., i+j);
    }
  }
  return tmp;  
}

Matrix Matrix::tr() const {
  Matrix tmp;
  for( int i=0; i<4; i++ ) {
    for( int j=0; j<4; j++ ) {
      tmp.x[j][i] = x[i][j];
    }
  }
  return tmp;  
}


Matrix translateMatrix(Vector a) {
  Matrix tmp;
  for(int i=0; i<3; i++)
    tmp(i,3) = a[i]; 
  return tmp;
}


Matrix scaleMatrix(float a) {
  Matrix tmp;
  for(int i=0; i<3; i++)
    tmp(i,i) = a;
  tmp(3,3)=1;
  return tmp;
}

Matrix scaleMatrix(Vector a) {
  Matrix tmp;
  for(int i=0; i<3; i++)
    tmp(i,i) = a[i];
  tmp(3,3) = 1;
  return tmp;
}


Matrix rotateMatrix(int axis, float angle) {
  Matrix tmp;
  int ax1 = (axis+1) % 3;
  int ax2 = (ax1+1) % 3;
  tmp(ax1,ax1) = cos(-angle);
  tmp(ax2,ax2) = tmp(ax1,ax1);
  tmp(ax1,ax2) = sin(-angle);
  tmp(ax2,ax1) = - tmp(ax1,ax2);

  return tmp;
}

Matrix rotateMatrix(Vector axis, float angle) {
  Matrix tmp;
  float ux = axis[0];
  float uy = axis[1];
  float uz = axis[2];
  float c = cos(angle);
  float s = sin(angle);

                          
  tmp(0,0) = c+(1-c)*ux*ux;
  tmp(0,1) = (1-c)*uy*ux-s*uz;
  tmp(0,2) = (1-c)*uz*ux+s*uy;

  tmp(1,0) = (1-c)*ux*uy+s*uz;
  tmp(1,1) = c+(1-c)*uy*uy;
  tmp(1,2) = (1-c)*uz*uy-s*ux;

  tmp(2,0) = (1-c)*ux*uz-s*uy;
  tmp(2,1) = (1-c)*uy*uz+s*ux;
  tmp(2,2) = c+(1-c)*uz*uz;
  
  return tmp;
}


Matrix frustumMatrix(float l, float r, float b, float t, float n, float f) {
    
    // Stolen from OpenGL documentation
  Matrix tmp;
  float invFmN = 1.0 / (f - n);
  float invTmB = 1.0 / (t - b);
  float invRmL = 1.0 / (r - l);
  
  tmp(0, 0) = 2.0*n*invRmL;
  tmp(0, 2) = (r+l)*invRmL;
  tmp(1, 1) = 2.0*n*invTmB;
  tmp(1, 2) = (t+b)*invTmB;
  tmp(2, 2) = -(f+n)*invFmN;
  tmp(2, 3) = -2.0*f*n*invFmN;
  tmp(3, 2) = -1;
  tmp(3, 3) = 0;
  /*
  float frustMatrixVals[4][4] = {
                            2.0*n*invRmL,       0,              (r+l)*invRmL,    0, 
                            0,                  2.0*n*invTmB,   (t+b)*invTmB,    0,
                            0,                  0,              -(f+n)*invFmN,   -2.0*f*n*invFmN,
                            0,                  0,              -1,              0
                          };
  
  return(Matrix(frustMatrixVals));
  */
  return tmp;
}

Matrix orthoMatrix(float l, float r, float b, float t, float n, float f) {
    
    // Stolen from OpenGL documentation
  Matrix tmp;
  float invFmN = 1.0 / (f - n);
  float invTmB = 1.0 / (t - b);
  float invRmL = 1.0 / (r - l);
  
  tmp(0, 0) = 2.0*invRmL;
  tmp(0, 3) = (-r-l)*invRmL;
  tmp(1, 1) = 2.0*invTmB;
  tmp(1, 3) = (-t-b)*invTmB;
  tmp(2, 2) = -2*invFmN;
  tmp(2, 3) = (-f-n)*invFmN;
  tmp(3, 3) = 1;
  return tmp;
}

  // Matrix that moves o1x1 and o1y1 to o2x2 and o2y2
Matrix coordChangeMatrix(Vector o1, Vector x1, Vector y1, Vector o2, Vector x2, Vector y2) {
  float x_angle = x2.angle(x1);
  Vector x_axis = (x2 * x1).norm();
  Matrix x_rotate = rotateMatrix(x_axis, x_angle);
  
  Vector int_y2 =  x_rotate * y2;    

  float y_angle = int_y2.angle(y1);
  Vector y_axis = (int_y2 * y1).norm();

  return  (rotateMatrix(y_axis, y_angle) * x_rotate) * translateMatrix(o1-o2);
}


Matrix distortMatrix(Vector ul, Vector ur, Vector bl, Vector br) {

    // perspective transformation code ruthlessly pirated from the Gimp (gimp.org)
      
  
  /* Determine the perspective transform that maps from
  * the unit cube to the transformed coordinates
  */
  
  float dx1, dx2, dx3, dy1, dy2, dy3;
  float matrix[4][4];
      
  dx1 = br[0] - ur[0];
  dx2 = ul[0] - ur[0];
  dx3 = bl[0] - br[0] + ur[0] - ul[0];

  dy1 = br[1] - ur[1];
  dy2 = ul[1] - ur[1];
  dy3 = bl[1] - br[1] + ur[1] - ul[1];

  /*  Is the mapping affine?  */
  if ((dx3 == 0.0) && (dy3 == 0.0)) {
    matrix[0][0] = br[0] - bl[0];
    matrix[0][1] = ur[0] - br[0];
    matrix[0][2] = bl[0];
    matrix[1][0] = br[1] - bl[1];
    matrix[1][1] = ur[1] - br[1];
    matrix[1][2] = bl[1];
    matrix[2][0] = 0.0;
    matrix[2][1] = 0.0;
  }
  else {
    float det1, det2;

    det1 = dx3 * dy2 - dy3 * dx2;
    det2 = dx1 * dy2 - dy1 * dx2;

    if (det1 == 0.0 && det2 == 0.0)
      matrix[2][0] = 1.0;
    else
      matrix[2][0] = det1 / det2;

    det1 = dx1 * dy3 - dy1 * dx3;

    if (det1 == 0.0 && det2 == 0.0)
      matrix[2][1] = 1.0;
    else
      matrix[2][1] = det1 / det2;

    matrix[0][0] = br[0] - bl[0] + matrix[2][0] * br[0];
    matrix[0][1] = ul[0] - bl[0] + matrix[2][1] * ul[0];
    matrix[0][2] = bl[0];

    matrix[1][0] = br[1] - bl[1] + matrix[2][0] * br[1];
    matrix[1][1] = ul[1] - bl[1] + matrix[2][1] * ul[1];
    matrix[1][2] = bl[1];
  }

  Matrix m = Matrix(matrix).tr();
  
  float dmVals[4][4] = {
    m(0, 0), m(1, 0),  0, m(2, 0), 
    m(0, 1), m(1, 1),  0, m(2, 1), 
          0,       0,  1,       0,
    m(0, 2), m(1, 2),  0,       1
  };
  return Matrix(dmVals);


}

Matrix realDistortMatrix(Vector ul, Vector ur, Vector bl, Vector br) {

    float m00, m01, m02, m10, m11, m12, m20, m21;
    float x00, x10, x20, x30, x01, x11, x21, x31;
    x00 = bl[0];
    x01 = bl[1];
    x10 = ul[0];
    x11 = ul[1];
    x20 = br[0];
    x21 = br[1];
    x30 = ur[0];
    x31 = ur[1];
        
m02 = ((x11 * x21 * x30 * ((2 * x11 * x20 - 2 * x10 * x21 - x11 * x30 + x21 * x30)) + 
  (x00 * x00) * ((x11 - x21)) * ((x11 - x31)) * ((x21 - x31)) - (((x11 * x11) * (x20 * x20) - (x10 * x10) * (x21 * x21) + 
  2 * x11 * (((-x10) + x20)) * x21 * x30)) * x31 + ((x11 * (x20 * x20) - (x10 * x10) * x21)) * (x31 * x31) + 
  2 * x00 * ((x11 * x20 - x10 * x21)) * ((x11 - x31)) * (((-x21) + x31)) + (x01 * x01) * ((x21 * (((x10 - x30)) * ((x10 - x30))) - 
  x11 * (((x20 - x30)) * ((x20 - x30))) - ((x10 - x20)) * ((x10 + x20 - 2 * x30)) * x31)) + x01 * (((-(x21 * x21)) * (((x10 - x30)) * ((x10 - x30))) + 
  (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) + 2 * ((x10 - x20)) * x21 * ((x00 - x30)) * x31 + ((x10 - x20)) * (((-2) * x00 + x10 + x20)) * (x31 * x31) + 
  2 * x11 * ((x10 - x20)) * ((x00 - x30)) * (((-x21) + x31))))))/((2 * x00 * ((x11 - x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - 
  x20 * x31)) + ((x11 * x20 - x10 * x21)) * ((x11 * x30 - x10 * x31)) * (((-x21) * x30 + x20 * x31)) + 
  x01 * (((-2) * ((x10 - x20)) * (((-x11) * x30 + x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x00 * (((x21 * x21) * (((x10 - x30)) * ((x10 - x30))) - 
  (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) - 2 * x10 * ((x10 - x20)) * x21 * x31 + ((x10 - x20)) * ((x10 + x20)) * (x31 * x31) + 
  2 * x11 * ((x10 - x20)) * ((x21 * x30 - x20 * x31)))))) + (x01 * x01) * ((x20 * x30 * (((-x21) * x30 + x20 * x31)) - 
  (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) + x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + 
  (x00 * x00) * ((x21 * x31 * (((-x21) * x30 + x20 * x31)) + (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) - 
  x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))); 


m12 = (((x00 * x00) * (x11 * x11) * x20 - x00 * (x11 * x11) * (x20 * x20) - 
  (x00 * x00) * x10 * (x21 * x21) + x00 * (x10 * x10) * (x21 * x21) + 
  (x01 * x01) * ((x10 - x20)) * ((x10 - x30)) * ((x20 - x30)) + 2 * x01 * ((x11 * x20 - x10 * x21 + 
  x00 * (((-x11) + x21)))) * ((x10 - x30)) * ((x20 - x30)) - (x00 * x00) * (x11 * x11) * x30 + (x11 * x11) * (x20 * x20) * x30 + 
  (x00 * x00) * (x21 * x21) * x30 - (x10 * x10) * (x21 * x21) * x30 + x00 * (x11 * x11) * (x30 * x30) - (x11 * x11) * x20 * (x30 * x30) - 
  x00 * (x21 * x21) * (x30 * x30) + x10 * (x21 * x21) * (x30 * x30) + 2 * ((x00 - x10)) * ((x00 - x20)) * ((x21 * ((x10 - x30)) + 
  x11 * (((-x20) + x30)))) * x31 + (((-x00) + x10)) * (((-x00) + x20)) * (((-x10) + x20)) * (x31 * x31)))/((2 * x00 * (((-x11) + 
  x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - x20 * x31)) - ((x11 * x20 - x10 * x21)) * ((x11 * x30 - x10 * x31)) * (((-x21) * x30 + x20 * x31)) + 
  (x01 * x01) * ((x20 * x30 * ((x21 * x30 - x20 * x31)) + (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) - x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 
  2 * x20 * ((x21 * x30 - x20 * x31)))))) + x01 * ((2 * ((x10 - x20)) * (((-x11) * x30 + x10 * x31)) * (((-x21) * x30 + x20 * x31)) + 
  x00 * (((-(x21 * x21)) * (((x10 - x30)) * ((x10 - x30))) + (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) + 2 * x10 * ((x10 - x20)) * x21 * x31 - ((x10 - 
  x20)) * ((x10 + x20)) * (x31 * x31) + 2 * x11 * ((x10 - x20)) * (((-x21) * x30 + x20 * x31)))))) + (x00 * x00) * ((x21 * x31 * ((x21 * x30 - x20 * x31)) - 
  (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) + x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))); 


m00 = (-(((((x01 - x11)) * ((x01 * ((x20 - x30)) + 
  x21 * x30 - x20 * x31 + x00 * (((-x21) + x31)))) * ((x11 * ((x20 - x30)) + x21 * x30 - x20 * x31 + 
  x10 * (((-x21) + x31))))))/((2 * x00 * ((x11 - x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - x20 * x31)) + ((x11 * x20 - x10 * x21)) * ((x11 * x30 - 
  x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x01 * (((-2) * ((x10 - x20)) * (((-x11) * x30 + x10 * x31)) * (((-x21) * x30 + x20 * x31)) + 
  x00 * (((x21 * x21) * (((x10 - x30)) * ((x10 - x30))) - (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) - 2 * x10 * ((x10 - x20)) * x21 * x31 + ((x10 - 
  x20)) * ((x10 + x20)) * (x31 * x31) + 2 * x11 * ((x10 - x20)) * ((x21 * x30 - x20 * x31)))))) + (x01 * x01) * ((x20 * x30 * (((-x21) * x30 + x20 * x31)) - 
  (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) + x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + 
  (x00 * x00) * ((x21 * x31 * (((-x21) * x30 + x20 * x31)) + (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) - 
  x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))))); 
  

m10 = (-(((((x00 - x10)) * (((-x21) * x30 + 
  x01 * (((-x20) + x30)) + x00 * ((x21 - x31)) + x20 * x31)) * (((-x21) * x30 + x11 * (((-x20) + x30)) + x10 * ((x21 - x31)) + 
  x20 * x31))))/((2 * x00 * (((-x11) + x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - x20 * x31)) - ((x11 * x20 - x10 * x21)) * ((x11 * x30 - 
  x10 * x31)) * (((-x21) * x30 + x20 * x31)) + (x01 * x01) * ((x20 * x30 * ((x21 * x30 - x20 * x31)) + (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) - 
  x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + x01 * ((2 * ((x10 - x20)) * (((-x11) * x30 + x10 * x31)) * (((-x21) * x30 + x20 * x31)) + 
  x00 * (((-(x21 * x21)) * (((x10 - x30)) * ((x10 - x30))) + (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) + 2 * x10 * ((x10 - x20)) * x21 * x31 - ((x10 - x20)) * ((x10 + x20)) * (x31 * x31) + 
  2 * x11 * ((x10 - x20)) * (((-x21) * x30 + x20 * x31)))))) + (x00 * x00) * ((x21 * x31 * ((x21 * x30 - x20 * x31)) - (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) + 
  x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))))); 


m20 = ((((x01 * x10 - x00 * x11)) * (((-x21) * x30 + x11 * (((-x20) + x30)) + x10 * ((x21 - x31)) + 
  x20 * x31)) * ((x01 * ((x20 - x30)) + x21 * x30 - x20 * x31 + x00 * (((-x21) + x31))))))/((2 * x00 * (((-x11) + x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - 
  x20 * x31)) - ((x11 * x20 - x10 * x21)) * ((x11 * x30 - x10 * x31)) * (((-x21) * x30 + x20 * x31)) + (x01 * x01) * ((x20 * x30 * ((x21 * x30 - x20 * x31)) + 
  (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) - x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + x01 * ((2 * ((x10 - x20)) * (((-x11) * x30 + 
  x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x00 * (((-(x21 * x21)) * (((x10 - x30)) * ((x10 - x30))) + (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) + 2 * x10 * ((x10 - x20)) * x21 * x31 - ((x10 - 
  x20)) * ((x10 + x20)) * (x31 * x31) + 2 * x11 * ((x10 - x20)) * (((-x21) * x30 + x20 * x31)))))) + (x00 * x00) * ((x21 * x31 * ((x21 * x30 - x20 * x31)) - 
  (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) + x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))); 


m01 = (-(((((x01 - x21)) * (((-x11) * x30 + x01 * (((-x10) + x30)) + x00 * ((x11 - x31)) + x10 * x31)) * ((x11 * ((x20 - x30)) + x21 * x30 - 
  x20 * x31 + x10 * (((-x21) + x31))))))/((2 * x00 * (((-x11) + x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - x20 * x31)) - ((x11 * x20 - x10 * x21)) * ((x11 * x30 - 
  x10 * x31)) * (((-x21) * x30 + x20 * x31)) + (x01 * x01) * ((x20 * x30 * ((x21 * x30 - x20 * x31)) + (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) - 
  x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + x01 * ((2 * ((x10 - x20)) * (((-x11) * x30 + x10 * x31)) * (((-x21) * x30 + x20 * x31)) + 
  x00 * (((-(x21 * x21)) * (((x10 - x30)) * ((x10 - x30))) + (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) + 2 * x10 * ((x10 - x20)) * x21 * x31 - ((x10 - x20)) * ((x10 + x20)) * (x31 * x31) + 
  2 * x11 * ((x10 - x20)) * (((-x21) * x30 + x20 * x31)))))) + (x00 * x00) * ((x21 * x31 * ((x21 * x30 - x20 * x31)) - (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) + 
  x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))))); 


m11 = (-(((((x00 - x20)) * (((-x11) * x30 + x01 * (((-x10) + x30)) + x00 * ((x11 - x31)) + x10 * x31)) * ((x11 * ((x20 - x30)) + x21 * x30 - 
  x20 * x31 + x10 * (((-x21) + x31))))))/((2 * x00 * ((x11 - x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - x20 * x31)) + ((x11 * x20 - x10 * x21)) * ((x11 * x30 - 
  x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x01 * (((-2) * ((x10 - x20)) * (((-x11) * x30 + x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x00 * (((x21 * x21) * (((x10 - x30)) * ((x10 - x30))) - 
  (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) - 2 * x10 * ((x10 - x20)) * x21 * x31 + ((x10 - x20)) * ((x10 + x20)) * (x31 * x31) + 2 * x11 * ((x10 - x20)) * ((x21 * x30 - 
  x20 * x31)))))) + (x01 * x01) * ((x20 * x30 * (((-x21) * x30 + x20 * x31)) - (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) + 
  x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + (x00 * x00) * ((x21 * x31 * (((-x21) * x30 + x20 * x31)) + 
  (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) - x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 2 * x21 * (((-x21) * x30 + x20 * x31)))))))))); 


m21 = ((((x01 * x20 - x00 * x21)) * ((x01 * ((x10 - x30)) + x11 * x30 - x10 * x31 + x00 * (((-x11) + x31)))) * ((x11 * ((x20 - x30)) + 
  x21 * x30 - x20 * x31 + x10 * (((-x21) + x31))))))/((2 * x00 * ((x11 - x21)) * ((x11 * x30 - x10 * x31)) * ((x21 * x30 - 
  x20 * x31)) + ((x11 * x20 - x10 * x21)) * ((x11 * x30 - x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x01 * (((-2) * ((x10 - x20)) * (((-x11) * x30 + 
  x10 * x31)) * (((-x21) * x30 + x20 * x31)) + x00 * (((x21 * x21) * (((x10 - x30)) * ((x10 - x30))) - (x11 * x11) * (((x20 - x30)) * ((x20 - x30))) - 2 * x10 * ((x10 - x20)) * x21 * x31 + ((x10 - 
  x20)) * ((x10 + x20)) * (x31 * x31) + 2 * x11 * ((x10 - x20)) * ((x21 * x30 - x20 * x31)))))) + (x01 * x01) * ((x20 * x30 * (((-x21) * x30 + x20 * x31)) - 
  (x10 * x10) * ((x20 * ((x21 - 2 * x31)) + x30 * x31)) + x10 * ((x11 * (((x20 - x30)) * ((x20 - x30))) + 2 * x20 * ((x21 * x30 - x20 * x31)))))) + 
  (x00 * x00) * ((x21 * x31 * (((-x21) * x30 + x20 * x31)) + (x11 * x11) * ((x20 * x21 + x30 * (((-2) * x21 + x31)))) - x11 * ((x10 * (((x21 - x31)) * ((x21 - x31))) + 
  2 * x21 * (((-x21) * x30 + x20 * x31))))))));
  
  

  
    float m[4][4] = {
      m00, m10, 0, m20,
      m01, m11, 0, m21,
        0,   0, 1,   0,
      m02, m12, 0,   1
    };
    
    Matrix dm = Matrix(m);
  
  return Matrix(dm);


}

